import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

public class HeapPQ<Key> implements PriorityQueue<Key> {
  private List<Key> pq;  // Heap ordered list
  private Comparator<Key> cmp;

  HeapPQ(Comparator<Key> cmp) {
    pq = new ArrayList<>();
    pq.add(null);
    this.cmp = cmp;
  }

  private void swim(int k) {
    while (k > 1 && !less(k / 2, k)) {
      swap(k / 2, k);
      k = k / 2;
    }
  }

  private void sink(int k) {
    while (2 * k <= size()) {
      int j = 2 * k;
      if (j < size() && !less(j, j + 1)) j++;
      if (less(k, j)) break;
      swap(k, j);
      k = j;
    }
  }

  private boolean less(int i, int j) {
    return cmp.compare(pq.get(i), pq.get(j)) < 0;
  }

  private void swap(int i, int j) {
    Key temp = pq.get(i);
    pq.set(i, pq.get(j));
    pq.set(j, temp);
  }

  @Override
  public void insert(Key v) {  // O(log n)
    pq.add(v);
    swim(size());
    assert invariant();
  }

  @Override
  public Key peek() {  // O(1)
    return pq.get(1);
  }

  @Override
  public Key pop() { // O(log n)
    Key min = peek();
    swap(1, size());
    pq.remove(size());
    sink(1);

    assert invariant();

    return min;
  }

  @Override
  public boolean isEmpty() {
    return pq.size() == 1;
  }

  @Override
  public int size() {
    return pq.size() - 1;
  }

  private boolean invariant() {
    for (int i = 1; i < size() / 2; i++) {
      Key x = pq.get(i);
      Key l = pq.get(2 * i + 1);
      Key r = pq.get(2 * i + 2);

      if (cmp.compare(x, l) > 0 || cmp.compare(x, r) > 0)
        return false;
    }
    return true;
  }
}
